module RuboCop
  module Cop
    module Betterment
      class AuthorizationInController < Cop
        attr_accessor :unsafe_parameters, :unsafe_regex

        # MSG_UNSAFE_CREATE = 'Model created/updated using unsafe parameters'.freeze
        MSG_UNSAFE_CREATE = <<~MSG.freeze
          Model created/updated using unsafe parameters.
          Please query for the associated record in a way that enforces authorization (e.g. "trust-root chaining"),
          and then pass the resulting object into your model instead of the unsafe parameter.

          INSTEAD OF THIS:
          post_parameters = params.permit(:album_id, :caption)
          Post.new(post_parameters)

          DO THIS:
          album = current_user.albums.find(params[:album_id])
          post_parameters = params.permit(:caption).merge(album: album)
          Post.new(post_parameters)

          See here for more information on this error:
          https://github.com/Betterment/rubocop-betterment/blob/master/README.md#bettermentauthorizationincontroller
        MSG

        def_node_matcher :model_new?, <<-PATTERN
          (send (const ... _) {:new :build :create :create! :find_or_create_by :find_or_create_by! :find_or_initialize_by :find_or_initialize_by!} ...)
        PATTERN

        def_node_matcher :model_update?, <<-PATTERN
          (send (...) {:assign_attributes :update :update! :find_or_create_by :find_or_create_by! :find_or_initialize_by :find_or_initialize_by! :update_attribute :update_attributes :update_attributes! :update_all :update_column :update_columns} ...)
        PATTERN

        def initialize(config = nil, options = nil)
          super(config, options)
          config = @config.for_cop(self)
          @unsafe_parameters = config.fetch("unsafe_parameters", []).map(&:to_sym)
          @unsafe_regex = Regexp.new config.fetch("unsafe_regex", ".*_id$")
          @wrapper_methods = {}
          @wrapper_names = []
        end

        def on_class(node)
          track_methods(node)
          track_assignments(node)
        end

        def on_send(node) # rubocop:disable Metrics/AbcSize,Metrics/PerceivedComplexity
          _receiver_node, _method_name, *arg_nodes = *node

          return if !model_new?(node) && !model_update?(node)

          arg_nodes.each do |argument|
            if argument.send_type?
              tag_unsafe_param_hash(argument)
              tag_unsafe_param_permit_wrapper(argument)
            elsif argument.variable?
              tag_unsafe_param_permit_wrapper(argument)
            elsif argument.hash_type?
              argument.children.each do |pair|
                next if pair.type != :pair

                _key, value = *pair.children
                tag_unsafe_param_hash(value)
                tag_unsafe_param_permit_wrapper(value)
              end
            end
          end
        end

        private

        def tag_unsafe_param_hash(node)
          unsafe_param = extract_parameter(node)
          add_offense(unsafe_param, message: MSG_UNSAFE_CREATE) if unsafe_param
        end

        def tag_unsafe_param_permit_wrapper(node)
          return if !node.send_type? && !node.variable?
          return if node.send_type? && (node.method_name == :[])

          name = Utils::Parser.get_root_token(node)
          add_offense(node, message: MSG_UNSAFE_CREATE) if @wrapper_names.include?(name)
        end

        # if a method returns params[...] or params.permit(...) then it will
        # be tracked; if the method does not explicitly return a params
        # sourced argument, then it will not be tracked
        def track_methods(node)
          methods = get_all_methods(node)

          methods.map do |method|
            method_returns = Utils::Parser.get_return_values(method)

            unless get_param_wrappers(method_returns).empty?
              @wrapper_methods[method.method_name] = method
              @wrapper_names << method.method_name
            end
          end
        end

        # keep track of all assignments that hold parameters
        def track_assignments(node)
          get_all_assignments(node).each do |assignment|
            variable_name, value = *assignment

            # if rhs calls params.permit, eg
            # - @var = params.permit(...)
            rhs_wrapper = get_param_wrappers([value])
            unless rhs_wrapper.empty?
              @wrapper_methods[variable_name] = rhs_wrapper[0]
              @wrapper_names << variable_name
            end

            # if rhs is a call to a parameter wrapper, eg
            # @var = parameter_wrapper
            root_token = Utils::Parser.get_root_token(value)
            if root_token && @wrapper_names.include?(root_token)
              @wrapper_methods[variable_name] = assignment
              @wrapper_names << variable_name
            end

            # if rhs extracts a parameter, eg
            # @var = params[:user_id]
            if extract_parameter(value)
              @wrapper_methods[variable_name] = assignment
              @wrapper_names << variable_name
            end
          end
        end

        def get_all_assignments(node)
          return [] unless node.children && node.class_type?

          node.descendants.select do |descendant|
            _lhs, rhs = *descendant
            descendant.equals_asgn? && (descendant.type != :casgn) && rhs&.send_type?
          end
        end

        def param_symbol?(name)
          name == :params
        end

        def get_all_methods(node)
          return [] unless node.children && node.class_type?

          node.descendants.select do |descendant|
            descendant.def_type?
          end
        end

        # this finds all calls to any method on a params like object
        # then walks up to find calls to permit
        # if the arguments to permit are "suspicious", then we add
        # the whole method to a list of methods that wrap params.permit
        def get_param_wrappers(methods) # rubocop:disable Metrics/AbcSize,Metrics/PerceivedComplexity
          wrappers = []

          methods.each do |method|
            return [] unless method.def_type? || method.send_type?

            method.descendants.each do |child|
              next unless child.send_type?
              next unless param_symbol?(child.method_name)

              child.ancestors.each do |ancestor|
                _receiver_node, method_name, *arg_nodes = *ancestor
                next unless ancestor.send_type?
                next unless method_name == :permit

                # we're only interested in calls to params.....permit(...)
                wrappers << method if contains_id_param?(arg_nodes)
              end
            end
          end

          wrappers
        end

        def sym_or_str?(arg)
          %i(sym str).include?(arg.type)
        end

        def array_or_hash?(arg)
          %i(array hash).include?(arg.type)
        end

        def contains_id_param?(arg_nodes)
          arg_nodes.any? do |arg|
            sym_or_str?(arg) && suspicious_id?(arg.value) ||
              array_or_hash?(arg) && contains_id_param?(arg.values)
          end
        end

        # check a symbol name against the cop's config parameters
        def suspicious_id?(symbol_name)
          @unsafe_parameters.include?(symbol_name.to_sym) || @unsafe_regex.match(symbol_name) # symbol_name.to_s.end_with?("_id")
        end

        def extract_parameter(argument)
          _receiver_node, method_name, *arg_nodes = *argument
          return unless argument.send_type? && method_name == :[]

          argument_name = Utils::Parser.get_root_token(argument)

          if param_symbol?(argument_name) || @wrapper_names.include?(argument_name)
            arg_nodes.find do |arg|
              sym_or_str?(arg) && suspicious_id?(arg.value)
            end
          end
        end
      end
    end
  end
end
